@@ -1,5 +1,5 @@
 """
-Copyright 2020 RICHARD TJÖRNHAMMAR
+Copyright 2021 RICHARD TJÖRNHAMMAR
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -47,6 +47,28 @@ def whiten_data ( Xdf ) :
     X_white = np.dot(X,np.dot( np.diag(s**-1),np.abs(v) )) # we don't know the sign
     return ( pd.DataFrame( X_white,index=Xdf.index.values,columns=Xdf.columns ) )
 
+def threshold ( E , A ) :
+    if not 'pandas' in str(type(A)) or not 'pandas' in str(type(E)):
+        print ( "ERROR MUST SUPPLY TWO PANDAS DATAFRAMES" )
+        return ( -1 )
+    thresholds_df = pd .DataFrame ( np.dot( E,A.T ) ,
+                          columns = A .index ,
+                          index   = E .index ) .apply ( lambda x:x/np.sum(E,1) )
+    return ( thresholds_df )
+
+def solve ( C = pd.DataFrame([ [10,1],[3,5] ]) ,
+            E = pd.DataFrame([ [25],[31] ]) ):
+    if not 'pandas' in str(type(C)) or not 'pandas' in str(type(E)):
+        print ( "ERROR MUST SUPPLY TWO PANDAS DATAFRAMES" )
+        return ( -1 )
+    recover = lambda U,S,Vt : np.dot(U*S,Vt)
+    cU, cS, cVt = np.linalg.svd(C, full_matrices=False )
+    cST   = 1/cS
+    psuedo_inverse = pd.DataFrame( recover(cVt.T,cST,cU.T) , index=C.columns ,columns=C.index )
+    identity  = np.dot(C,psuedo_inverse)
+    TOLERANCE = np.max( np.sqrt( ( identity * ( ( 1-np.eye(len(np.diag(identity)))) ) )**2 ))
+    return ( np.dot( psuedo_inverse,E),TOLERANCE )
+
 import re
 def find_category_variables( istr ) :
     return ( re.findall( r'C\((.*?)\)', istr ) )
@@ -95,7 +117,7 @@ def quantify_density_probability ( rpoints , cutoff = None ) :
     if not cutoff is None :
         resolution = 10. ; nbins = 100.
         #
-        # ONLY FOR ASSESING 
+        # ONLY FOR ASSESING
         h1,h2      = np.histogram(rpoints,bins=int(np.ceil(len(rpoints)/resolution)))
         bin_radius = 0.5 * ( h2[1:] + h2[:-1] )
         radial_density = np.cumsum( h1 )/np.sum( h1 ) # lt
@@ -117,7 +139,7 @@ def interpret_problem ( analyte_df , journal_df , formula , bVerbose=False ) :
     # THE JOURNAL_DF IS THE COARSE GRAINED DATA (THE MODEL)
     # THE ANALYTE_DF IS THE   FINE GRAINED DATA  (THE DATA)
     # THE FORMULA IS THE SEMANTIC DESCRIPTION OF THE PROBLEM
-    # 
+    #
     interaction_pairs = find_category_interactions ( formula.split('~')[1] )
     add_pairs = []
     if len( interaction_pairs )>0 :
@@ -143,7 +165,7 @@ def interpret_problem ( analyte_df , journal_df , formula , bVerbose=False ) :
         if encoding_df is None :
             encoding_df = add_df.T
         else :
-            encoding_df = pd.concat([ encoding_df.T , 
+            encoding_df = pd.concat([ encoding_df.T ,
                             journal_df.loc[ [ c.replace(' ','') for c in formula.split('~')[1].split('+') if not 'C(' in c] , : ] ]).T
     return ( encoding_df )
 
@@ -167,12 +189,12 @@ def calculate_alignment_properties ( encoding_df , quantx, quanty, scorex,
         exit(1)
     #
     # THESE ARE THE CATEGORICAL DESCRIPTORS
-    use_centroid_indices = [ i for i in range(len(encoding_df.columns.values)) if ( 
-                             encoding_df.columns.values[i] not in set( exclude_labels_from_centroids ) 
+    use_centroid_indices = [ i for i in range(len(encoding_df.columns.values)) if (
+                             encoding_df.columns.values[i] not in set( exclude_labels_from_centroids )
                            ) ]
     #
     use_centroids = list(  quanty[use_centroid_indices]  )
-    use_labels    = list( encoding_df.columns.values[use_centroid_indices] )  
+    use_labels    = list( encoding_df.columns.values[use_centroid_indices] )
     #
     if owner_by == 'tesselation' :
         transcript_owner = [ use_labels[ np.argmin([ np.sum((xw-cent)**2) for cent in use_centroids ])] for xw in quantx ]
@@ -219,7 +241,7 @@ def calculate_alignment_properties ( encoding_df , quantx, quanty, scorex,
     # print ( 'THE EQUIDISTANT 1D STATS' )
     corresponding_pvalue , corresponding_density , corresponding_radius = quantify_density_probability ( rpoints , cutoff = blur_cutoff )
     #
-    # print ( 'THE TWO XY 1D STATS' ) 
+    # print ( 'THE TWO XY 1D STATS' )
     corr_pvalue_0 , corr_density_0 = quantify_density_probability ( xpoints )
     corr_pvalue_1 , corr_density_1 = quantify_density_probability ( ypoints )
     #
@@ -275,7 +297,7 @@ def run_rpls_regression ( analyte_df , journal_df , formula ,
                         ) :
     inventors__ = "Richard Tjörnhammar (RT) and Edward Tjörnhammar"
     NOTE__ = "Edward Tjörnhammar (ET) early major contributor to this method. Inventors: "+inventors__+". RT is the main developer."
-    
+
     encoding_df = interpret_problem ( analyte_df , journal_df , formula , bVerbose = bVerbose )
     from sklearn.cross_decomposition import PLSRegression as PLS
 
@@ -292,7 +314,7 @@ def run_rpls_regression ( analyte_df , journal_df , formula ,
     			journal_df = journal_df, analyte_df = analyte_df , blur_cutoff = blur_cutoff ,
     			bVerbose = bVerbose, exclude_labels_from_centroids = exclude_labels_from_centroids ,
 			study_axii = study_axii , owner_by = owner_by )
-    
+
     return ( res_df )
 
 
@@ -329,7 +351,7 @@ def run_shape_alignment_clustering ( analyte_df , journal_df , formula, bVerbose
 	return ( res_df , clusters_df )
 
 def knn_clustering_alignment( P , Q ) :
-    
+
     NOTE_ = "This is just a standard kmeans in arbitrary dimensions that start out with centroids that have been shape aligned"
     ispanda = lambda P: 'pandas' in str(type(P)).lower()
     BustedPanda = lambda R : R.values if ispanda(R) else R
@@ -340,7 +362,7 @@ def knn_clustering_alignment( P , Q ) :
                     bReturnTransform = False ,
                     bShiftModel      = True  ,
                     bUnrestricted    = True  )
-    
+
     if ispanda ( Q ) :
         #
         # FOR DIAGNOSTIC PURPOSES
@@ -388,7 +410,7 @@ def run_shape_alignment_regression( analyte_df , journal_df , formula ,
 	centroids_df = pd.DataFrame ( centroids ,
 			index = encoding_df.columns ,
 			columns = encoding_df.index )
-			
+
 	xws = ifit.WeightsAndScoresOf( P )
 	yws = ifit.WeightsAndScoresOf( centroids )
 
@@ -402,7 +424,7 @@ def run_shape_alignment_regression( analyte_df , journal_df , formula ,
 			analyte_df = analyte_df.copy() , journal_df = journal_df.copy() ,
 			blur_cutoff = blur_cutoff , bVerbose = bVerbose,
 			exclude_labels_from_centroids = exclude_labels_from_centroids ,
-			study_axii = study_axii , owner_by = owner_by, synonyms=synonyms )		
+			study_axii = study_axii , owner_by = owner_by, synonyms=synonyms )
 	return ( res_df )
 
 
@@ -1012,7 +1034,7 @@ def retrieve_genes_of ( group_name, grouping_file, delimiter='\t', identifier='E
             if group_name is None :
                 line  = line_.replace( '\n','' )
                 dline = line.split( delimiter )
-                if identifier is None : 
+                if identifier is None :
                     [ all_analytes.append(d) for d in dline ]
                 else :
                     [ all_analytes.append(d) for d in dline if identifier in d ]
@@ -1020,7 +1042,7 @@ def retrieve_genes_of ( group_name, grouping_file, delimiter='\t', identifier='E
                 if group_name in line_:
                     line=line_.replace('\n','')
                     dline = line.split(delimiter)
-                    if identifier is None : 
+                    if identifier is None :
                         return ( [ d for d in dline ] )
                     else:
                         return ( [ d for d in dline if identifier in d ] )
@@ -1068,8 +1090,15 @@ def add_kendalltau( analyte_results_df , journal_df , what='M' , sample_names =
 
 def calculate_rates( journal_df , inferred_df ,
                      formula , inference_label = 'owner',
-                     bVerbose=False ) :
+                     bVerbose = False ,
+                     strictness = 'intersect' ) :
 
+    strictness_function = { 'any':any,'intersect':lambda x:x }
+
+    if strictness not in set(strictness_function.keys()):
+        print ( 'ERROR: COULD NOT ASSIGN A STRICTNESS FUNCTION' )
+        print ( 'VIABLE STRICTNESS OPTIONS ARE: ' , set( strictness_function.keys()) )
+        exit(1)
     check = []
     for idx in journal_df.index.values:
         if idx in formula :
@@ -1079,16 +1108,16 @@ def calculate_rates( journal_df , inferred_df ,
         print( 'Cannot assert quality' )
         results_lookup = dict( )
         return ( results_lookup )
-   
+
     known_df = journal_df.loc[check,:]
-    
+
     known_OH = create_encoding_journal( known_df.index.values, known_df )
     known_instances = known_OH.index
     inferred_OH = known_OH*0
     for o in inferred_df.iterrows():
         for known_instance in known_instances:
             inferred_OH.loc[known_instance,o[0]] = int( known_instance in o[1][inference_label] )
-            
+
     OH_not = lambda df:df.apply( lambda x:1-x )
 
     not_known_OH    = OH_not( known_OH )
@@ -1100,19 +1129,19 @@ def calculate_rates( journal_df , inferred_df ,
         print(inferred_OH.iloc[:6,:6])
         print(np.sum(np.sum(inferred_OH.iloc[:6,:6] * known_OH.iloc[:6,:6] )) )
 
-    TP = np.sum( np.sum( inferred_OH     * known_OH     ) )
-    FP = np.sum( np.sum( inferred_OH     * not_known_OH ) )
-    TN = np.sum( np.sum( not_inferred_OH * not_known_OH ) )
-    FN = np.sum( np.sum( not_inferred_OH * known_OH     ) )
+    TP = np.sum( np.sum( ( inferred_OH     * known_OH     ).apply(strictness_function[strictness]) ) )
+    FP = np.sum( np.sum( ( inferred_OH     * not_known_OH ).apply(strictness_function[strictness]) ) )
+    TN = np.sum( np.sum( ( not_inferred_OH * not_known_OH ).apply(strictness_function[strictness]) ) )
+    FN = np.sum( np.sum( ( not_inferred_OH * known_OH     ).apply(strictness_function[strictness]) ) )
 
-    results_lookup = {  'TP':TP , 'TN':TN ,
-                'FN':FN , 'FP':FP ,
+    results_lookup = { 'TP':TP , 'TN':TN ,
+                'FN':FN ,'FP':FP ,
                 'sensitivity' : TP / ( TP+FN ) ,
                 'specificity' : TN / ( TN+FP ) ,
                 'precision'   : TP / ( TP+FP ) ,
                 'accuracy'    : ( TP+TN ) / ( TP+TN+FP+FN ) ,
                 'negation'    : TN / ( TN+FN ) , # FNR
-                'FPR'         : FP / ( FP+TN ) , # False positive rate
+                'FPR:'        : FP / ( FP+TN ) , # False positive rate
                 'FDR'         : FP / ( FP+TP )   # False discovery rate
     }
     return ( results_lookup )
@@ -1124,11 +1153,26 @@ def assign_quality_measures( journal_df , result_dfs ,
 
     for label in [ col for col in result_dfs[0].columns if plabel in col[-2:] ] :
         result_dfs[0].loc[:, label[:-2]+',q'] = [ qvs[0] for qvs in qvalues( result_dfs[0].loc[:,label].values ) ]
-   
+
     results_lookup = calculate_rates ( journal_df , result_dfs[1] ,
                           formula , inference_label = inference_label )
     return( results_lookup )
 
+import math
+def isItPrime( N , M=None,p=None,lM05=None ) :
+    if p is None :
+        p = 1
+    if M is None :
+        M = N
+    if lM05 is None:
+        lM05 = math.log(M)*0.5
+    if ((M%p)==0 and p>=2) :
+        return ( N==2 )
+    else :
+       if math.log(p) > lM05:
+           return ( True )
+       return ( isItPrime(N-1,M=M,p=p+1,lM05=lM05) )
+
 
 if __name__ == '__main__' :
 
